from .base_session import IosSession
from .ms_samlpr import SamlSession
try:
    # Optional Windows module that requires the clr package.
    from .sec_api import SecApiSession
except ImportError:
    pass



class IosError(OSError):
    """An ambiguous exception occurred during this operation."""


class IosAuthenticationError(IosError, ConnectionRefusedError):
    """Authentication error while connecting to IOS; check username and password."""


class IosConnectionError(IosError, ConnectionError):
    """Exception occurred during IOS connection process."""


class IosDataError(IosError, ValueError):
    """Received an error response from the IOS server."""



from base64 import b64encode
from io import IOBase
from logging import getLogger
from typing import Union
# noinspection PyPackageRequirements
from bs4 import BeautifulSoup
from requests import Session
from requests.compat import cookielib
from urllib.parse import urljoin, urlparse
from pylawson import IosAuthenticationError, IosConnectionError
from pylawson.client import IosSession

logger = getLogger(__name__)


class SamlSession(IosSession):
    def __init__(self, json_file: Union[str, IOBase] = None, lawson_server: str = None, ident_server: str = None,
                 username: str = None, password: str = None):
        super().__init__(json_file=json_file, lawson_server=lawson_server, ident_server=ident_server,
                         username=username, password=password)
        self._sso = None
        headers = {
            'Upgrade-Insecure-Requests': '1', 'Accept-Language': 'en-US,en',
            'Accept': 'text/html,application/xhtml+xml,application/xml,image/webp,*/*',
            'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) ' +
                          'Chrome/53.0.2785.143 Safari/537.36'
        }
        ip_host = self._params['ident_host'] or urlparse(self._params['lawson_server']).netloc
        ip_cookie = cookielib.Cookie(
            version=0, name='MSISIPSelectionPersistent',
            value=b64encode(self._params['ident_server'].encode('utf-8')).decode('utf-8'),
            port=None, port_specified=False, domain=ip_host, domain_specified=False,
            domain_initial_dot=False, path='/adfs/ls', path_specified=True, secure=True,
            expires=cookielib.timegm(cookielib.time.localtime()) + 2500000, discard=False,
            comment=None, comment_url=None, rest={'HttpOnly': None}, rfc2109=False)
        self.session = Session()
        self.session.cookies.set_cookie(ip_cookie)
        self.session.headers.update(headers)
        logger.debug('Basic SamlSession instantiation completed.')
        self._auth()

    def __bool__(self):
        soup = BeautifulSoup(self.get('?_action=PING'), 'html.parser')
        if soup.find('sessionstatus').text == 'true':
            status = 'Active as {}, '.format(soup.find('username').text)
            milliseconds = self._sso[1] + int(soup.find('time_remaining').text)
            status += 'time remaining: {}'.format(
                cookielib.datetime.timedelta(milliseconds=milliseconds).__str__()[:-5])
            logger.debug(msg=status)
            return True

        return False

    def get(self, url: str) -> str:
        url = urljoin(self._params['lawson_server'], url)
        return self.session.get(url=url).text

    def post(self, url: str, data: dict) -> str:
        url = urljoin(self._params['lawson_server'], url)
        return self.session.post(url=url, data=data).text

    def close(self):
        if self:
            self.get(url='?_action=LOGOUT')
            self.session.close()
        logger.info(msg='Closed session.')

    def _auth(self):
        """Perform series of requests for SAML authentication."""

        # 1 - Request the target resource at service provider - status 302
        response = self.session.get(url=self._params['lawson_server'], allow_redirects=False)
        logger.debug(msg='Auth #1 - request target resource')
        if 'wa=wsignin' not in response.headers.get('Location', '') or response.status_code != 302:
            msg = 'Unexpected response from initial Target Resource Request.'
            logger.error(msg=msg)
            raise IosConnectionError(msg)

        # 2 - Discover Identity Provider - form to select IdP (unless persistent cookie) - status 200
        id_response = self.session.get(url=response.headers['Location'], allow_redirects=False)
        logger.debug(msg='Auth #2 - discover identity provider')
        if id_response.status_code == 200:
            url, data = self._form(id_response)

            # 3 - Redirect to SSO Service at IdP - status 302
            response = self.session.post(url, allow_redirects=False, data=data)
            logger.debug(msg='Auth #3 - redirect to SSO service')
            redirect = response.headers.get('Location')
            if response.cookies.get('MSISIPSelectionSession') is None:
                msg = 'Unexpected response from Identity Provider Form Submission.'
                logger.error(msg=msg)
                raise IosConnectionError(msg)
        else:
            redirect = id_response.headers.get('Location')
            if id_response.cookies.get('MSISIPSelectionSession') is None:
                msg = 'Unexpected response from SSO Service Redirect.'
                logger.error(msg=msg)
                raise IosConnectionError(msg)

        # 4 - Request the SSO Service at IdP - provides sign in form - status 200
        response = self.session.get(redirect, allow_redirects=False, headers={'Referer': id_response.url})
        logger.debug(msg='Auth #4 - request SSO service (sign in form)')
        url, data = self._form(response)
        if 'wa=wsignin' not in url or response.status_code != 200:
            msg = 'Unexpected response from Sign In Form Submission.'
            logger.error(msg=msg)
            raise IosConnectionError(msg)

        # 5 - Identify the user - sets auth cookies - status 302
        response = self.session.post(url, allow_redirects=False, data=data, headers={'Referer': response.url})
        logger.debug(msg='Auth #5 - identify the user')
        if response.cookies.get('MSISAuth') is None or response.status_code != 302:
            msg = 'Invalid username or password. (Is password expired?)'
            logger.error(msg=msg)
            raise IosAuthenticationError(msg)

        # 6 - Respond with XHTML form - status 200 (JS autosubmit script)
        response = self.session.get(
            response.headers['Location'], allow_redirects=False, headers={'Referer': response.url}
        )
        logger.debug(msg='Auth #6 - XHTML form')
        url, data = self._form(response)
        if response.cookies.get('MSISAuthenticated') is None or response.status_code != 200:
            msg = 'Unexpected response from SSO XHTML Form.'
            logger.error(msg=msg)
            raise IosConnectionError(msg)

        # 7 - Request Assertion Consumer Service at Service Provider - status 200 (JS autosubmit script)
        response = self.session.post(url, allow_redirects=False, data=data, headers={'Referer': response.url})
        logger.debug(msg='Auth #7 - request assertion consumer service')
        url, data = self._form(response)
        if response.cookies.get('MSISSignOut') is None or response.status_code != 200:
            msg = 'Unexpected response from Assertion Consumer Service Request.'
            logger.error(msg=msg)
            raise IosConnectionError(msg)

        # 8 - Redirect to target resource - sets C.LWSN session cookie finally - status 302
        response = self.session.post(url, allow_redirects=False, data=data, headers={'Referer': response.url})
        logger.debug(msg='Auth #8 - redirect to target resource')
        self._sso = (
            response.headers.get('SSO_STATUS', 'NoStatusHeader'),  # LoginSuccessful
            int(response.headers.get('SSO_TIMEOUT_REMAINING', 0))  # 3600000
        )
        logger.info('SSO Status: {}, Timeout: {} '.format(self._sso[0], self._sso[1]))
        if response.cookies.get('C.LWSN') is None or response.status_code != 302:
            msg = 'Unexpected response from Target Resource Redirect.'
            logger.error(msg=msg)
            raise IosConnectionError(msg)

        # 9 - Request target resource again - status 200 - (JS redirects to LOGINCOMPLETE)
        response = self.session.get(
            response.headers['Location'], allow_redirects=False,
            headers={'Referer': response.url}
        )
        logger.debug(msg='Auth #9 - request target resource again')
        if 'LOGINCOMPLETE' not in response.text:
            msg = 'Unexpected response from Target Resource.'
            logger.error(msg=msg)
            raise IosConnectionError(msg)

        # Get Transfer Session URL for session refresh later.
        self._xfer_url = self.session.get(urljoin(self._params['lawson_server'], '?_action=GET_XFER_SESSION')).text
        logger.debug(msg='Auth complete; XFER_SESSION response: {}.'.format(self._xfer_url))

        # Get Profile attributes
        soup = BeautifulSoup(self.get('/servlet/Profile?section=attributes'), 'html.parser')
        self._profile.__dict__ = {element.attrs['name'].lower(): element.attrs['value'] for element in
                                  soup.find_all('attr')}
        logger.debug('Populated profile.')

    def _form(self, response) -> (str, dict):
        """Read Response text and return Form URL and Data to post."""
        soup = BeautifulSoup(response.text, 'html.parser')
        data = {}
        select = soup.find('select')
        if len(soup.find_all('form')) != 1:
            msg = 'Unsupported number of Forms on page [{}].'.format(response.url)
            logger.error(msg=msg)
            raise IosConnectionError(msg)
        if select is not None and 'Provider' in select.get('name'):
            data[select] = self._params['ident_server']
        for element in soup.find_all('input'):
            if element.has_attr('name') and element.has_attr('value'):
                data[element['name']] = element['value']
            elif element.has_attr('name') and 'Username' in element.get('name'):
                data[element['name']] = self._params['username']
            elif element.has_attr('name') and 'Password' in element.get('name'):
                data[element['name']] = self._params['password']
        action = urljoin(response.url, soup.find('form').get('action'))
        return action, data



from io import IOBase
import json
from typing import Union, Optional


class Profile:
    """Container for Lawson user profile attributes."""
    def __repr__(self):
        return self.__class__.__name__ + repr(self.__dict__)


class IosSession:
    """Base class for an Infor Lawson Connection session object."""
    def __init__(self, json_file: Union[str, IOBase] = None, lawson_server: str = None, ident_server: str = None,
                 username: str = None, password: str = None):
        self._profile = Profile()
        self._xfer_url = None
        self._params = {
            'lawson_server': lawson_server,
            'ident_server': ident_server,
            'ident_host': None,
            'username': username,
            'password': password
        }
        if json_file:
            fp = json_file if isinstance(json_file, IOBase) else open(json_file)
            json_data = json.load(fp=fp)
            self._params.update(json_data.get('lawson', {}))

    def __repr__(self):
        return self.__class__.__name__

    def __bool__(self):
        return False

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    @property
    def is_authenticated(self):
        return self.__bool__()

    @property
    def profile(self) -> Profile:
        return self._profile

    def get(self, url: str):
        raise NotImplementedError

    def post(self, url: str, data: dict):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    def _generic_call(self, url: str, data: dict, productline_key: Optional[str]) -> str:
        """Wraps self.post to send a specific action with product line."""
        call_data = dict()
        if productline_key:
            # noinspection PyUnresolvedReferences
            call_data[productline_key] = self.profile.productline
        call_data.update(data)
        if not call_data:
            return self.get(url=url)
        return self.post(url=url, data=call_data)

    def tokens(self, data: dict):
        """Lawson ListTokens Action."""
        url = '/lawson-ios/action/ListTokens'
        productline_key = 'productLine'
        return self._generic_call(url=url, data=data, productline_key=productline_key)

    def attachments(self, data: dict):
        """Lawson ListAttachments Action."""
        url = '/lawson-ios/action/ListAttachments'
        productline_key = 'dataArea'
        return self._generic_call(url=url, data=data, productline_key=productline_key)

    def data(self, data: dict):
        """Lawson Data call."""
        url = '/servlet/Router/Data/erp'
        productline_key = 'PROD'
        return self._generic_call(url=url, data=data, productline_key=productline_key)

    def drill(self, data: dict):
        """Lawson Drill call."""
        url = '/servlet/Router/Drill/erp'
        productline_key = 'PROD'
        return self._generic_call(url=url, data=data, productline_key=productline_key)

    def transaction(self, data: dict):
        """Lawson Transaction call."""
        url = '/servlet/Router/Transaction/erp'
        productline_key = '_PDL'
        return self._generic_call(url=url, data=data, productline_key=productline_key)

    def what(self, data: dict):
        """Lawson What call."""
        url = '/servlet/What'
        return self._generic_call(url=url, data=data)



"""SecApiSession: IosSession using Infor Lawson Office Add-ins .NET library (Infor Security sec-api.dll)."""
from io import IOBase
from logging import getLogger
import os
from typing import Union
from urllib.parse import urlencode, urljoin
# noinspection PyPackageRequirements
from bs4 import BeautifulSoup
from pylawson import IosError, IosAuthenticationError, IosConnectionError
from pylawson.client import IosSession

# noinspection PyPackageRequirements
import clr
try:
    _ = clr.AddReference(
        os.getenv('PROGRAMFILES(x86)', r'c:\Program Files (x86)') + r'\Lawson Software\Office Add-ins\sec-api.dll'
    )
except FileNotFoundError:
    _ = clr.AddReference(
        os.getenv('PROGRAMFILES', r'c:\Program Files') + r'\Lawson Software\Office Add-ins\sec-api.dll'
    )
# noinspection PyUnresolvedReferences,PyPackageRequirements
from Security import EnvironmentManagement, Factory, LawsonHttpClient, LSConnect
# noinspection PyUnresolvedReferences,PyPackageRequirements
from System import Threading

logger = getLogger(__name__)


def authorized(func):
    """Wrapper to ensure authorization session is current before making a data call."""
    def wrapper(*args):
        instance = args[0]
        if not instance.server or not instance.connection.IsAuthenticated():
            instance.login(login_string='Log in again to continue.')
        try:
            assert instance.profile.productline is not None
        except (AttributeError, AssertionError):
            instance.login(login_string='Log in again to continue.')
        return func(*args)

    return wrapper


class SecApiSession(IosSession):
    """
    Subclasses pylawson.IosSession using Infor Lawson Office Add-ins .NET library (Infor Security sec-api.dll)

    The Infor sec-api library will pop up a login window; as such, this class does not accept username/password.
    """
    def __init__(self, json_file: Union[str, IOBase] = None, lawson_server: str = None, ident_server: str = None):
        super().__init__(json_file=json_file, lawson_server=lawson_server, ident_server=ident_server,
                         username=None, password=None)
        # The sec-api .NET library crashes if it doesn't detect a Single Apartment State environment.
        thread = Threading.Thread.CurrentThread
        if not thread.TrySetApartmentState(Threading.ApartmentState.STA):
            msg = 'CLR failed to set .NET Single Apartment State environment.'
            logger.error(msg=msg)
            raise IosError(msg)

        # The Security Factory class has the function to instantiate the Authenticator Interface,
        # which in turn has the Login function which returns our Connection Interface.
        factory = Factory.ClientCOMSecurityFactory()
        factory.Path = os.getcwd()
        self.authenticator = factory.GetClientSecurityAuthenticator()
        if factory.Error:
            msg = 'SEC-API error: {}'.format(factory.Error)
            logger.error(msg)
            raise IosConnectionError(msg)

        # The balance of our methods are found on the Connection Interface.
        self.connection = None
        self.server = None
        logger.debug('Basic SecApiSession instantiation completed.')

        # Run login, which populates self.connection, self._xfer_url, self.server, and self._profile
        self.login(clientDisplayName='Python Infor Login')

    def __bool__(self) -> bool:
        """Status of connection."""
        return self.connection.IsAuthenticated()

    def _get(self, url: str) -> str:
        """Send GET call to server, return response data. (Not wrapped, for initial profile call."""
        url = urljoin(self.server, url)
        response = self.connection.SendData(url)
        return response.ResponseData

    @authorized
    def get(self, url: str) -> str:
        """Send GET call to server, return response data."""
        return self._get(url=url)

    @authorized
    def post(self, url, data):
        """Mimic POST call to server with GET, return response data."""
        s = urlencode(data) if isinstance(data, dict) else data
        return self.get('?{}'.format(s))

    def close(self):
        """Logout and remove connection."""
        self.connection.Logout()
        self.connection = None
        logger.info(msg='Closed session.')

    # noinspection PyPep8Naming
    def login(self, clientDisplayName: str):
        """Pop up Login window for user to log in to Lawson.

        Creates the IConnectionHandler object, then populates the _xfer_url, server, and _profile variables.

        :param clientDisplayName: .NET window title
        :raise IosAuthenticationError: if login window returns without valid session
        """
        self.connection = self.authenticator.DoActiveClientLogin(clientDisplayName)
        if self:
            self._xfer_url = self.connection.GetTransferSessionToken()  # Transfer token for session persistence.
            self.server = self.connection.GetConnectedServerUrl()  # Lawson server which we have logged in to.
            # Profile Attributes for the logged-in user.
            soup = BeautifulSoup(self._get('/servlet/Profile?section=attributes'), 'html.parser')
            self._profile.__dict__ = {element.attrs['name'].lower(): element.attrs['value'] for element in
                                      soup.find_all('attr')}
            logger.debug('Populated profile.')

        else:
            msg = 'Authentication failed after login window returned.'
            logger.error(msg=msg)
            raise IosAuthenticationError(msg)



from logging import getLogger
# noinspection PyPackageRequirements
from bs4 import BeautifulSoup
from .client import IosSession as Session
from .exceptions import IosDataError

logger = getLogger(__name__)


class LawsonBase:
    """Base class for Lawson data objects."""
    def __init__(self, session: Session, **kwargs):
        self.session = session
        self._soup = None
        self._xml = None
        self.params = kwargs

    def __repr__(self):
        return self.__class__.__name__

    @property
    def xml(self):
        return self._xml

    @xml.setter
    def xml(self, value: str):
        self._soup = None
        self._xml = value

    @property
    def soup(self):
        if not self._soup:
            self._soup = BeautifulSoup(self.xml, 'html.parser')
        return self._soup

    def _error_check(self):
        if self.soup.contents[0].name == 'ERROR':
            msg = 'Infor error: [{}] {}'.format(self.soup.contents[0].attrs.get('key'), self.soup('MSG')[0].text)
            logger.error(msg=msg)
            raise IosDataError(msg)

    def query(self, **kwargs):
        raise NotImplementedError

    def upload(self, **kwargs):
        raise NotImplementedError


class What(LawsonBase):
    def query(self, jar: str='IOS.jar'):
        self.params.update({'_JAR': jar})
        # Expected response like:
        # <WHAT laversion="10.0.5.0.1093 2015-09-22 04:00:00"><JAR name="IOS.jar" path="..."><MANIFEST><![CDATA[
        # Implementation-Vendor: Lawson Software
        # Implementation-Title: IOS
        # Implementation-Version: 8-)@(#)@10.0.5.0.1093 2015-09-22 04:00:00]]></MANIFEST><LAVERSION>
        # <![CDATA[8-)@(#)@10.0.5.0.1093 2015-09-22 04:00:00]]></LAVERSION></JAR></WHAT>"""
        self.xml = self.session.what(data=self.params)
        self._error_check()
        return self

    upload = None


class Account(LawsonBase):
    def query(self):
        raise NotImplementedError

    def upload(self):
        raise NotImplementedError


class Activity(LawsonBase):
    def query(self):
        self.params.update({'FILE': 'ACACTIVITY', 'OUT': 'XML', 'NEXT': 'FALSE', 'keyUsage': 'PARAM'})
        self.xml = self.session.data(data=self.params)
        self._error_check()
        return self

    def upload(self):
        self.params.update({'_TKN': 'AC10.1', '_RTN': 'DATA', '_TDS': 'IGNORE', '_OUT': 'XML', '_EOT': 'TRUE'})
        self.xml = self.session.transaction(data=self.params)
        self._error_check()
        return self


class Journal(LawsonBase):
    query = None

    def upload(self):
        self.params.update({'_TKN': 'GL40.2', '_RTN': 'DATA', '_TDS': 'IGNORE', '_OUT': 'XML', '_EOT': 'TRUE'})
        self.xml = self.session.transaction(data=self.params)
        self._error_check()
        return self


class JournalLine(LawsonBase):
    def query(self):
        self.params.update(
            {'FILE': 'GLTRANS', 'INDEX': 'GLTSET3', 'OUT': 'XML', 'NEXT': 'FALSE', 'MAX': '10000', 'keyUsage': 'PARAM'})
        self.xml = self.session.data(data=self.params)
        self._error_check()
        return self

    def upload(self):
        self.params.update(
            {'_TKN': 'GL40.1', '_RTN': 'DATA', '_TDS': 'IGNORE', '_OUT': 'XML', '_EOT': 'TRUE', '_INITDTL': 'TRUE'})
        self.xml = self.session.transaction(data=self.params)
        self._error_check()
        return self


class InterfaceLine(LawsonBase):
    def query(self):
        raise NotImplementedError

    def upload(self):
        raise NotImplementedError
